package de.ozgcloud.formcycle;

import static de.ozgcloud.formcycle.ExecutionResultTestFactory.*;
import static de.ozgcloud.formcycle.OzgCloudPlugin.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.Properties;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import de.ozgcloud.formcycle.errorhandling.TechnicalException;
import de.xima.fc.exceptions.NodeThrewException;
import de.xima.fc.interfaces.workflow.execution.INormalCompletionResult;
import de.xima.fc.interfaces.workflow.params.INodeExecutionParams;
import de.xima.fc.interfaces.workflow.params.INormalCompletionResultBuilder;
import de.xima.fc.interfaces.workflow.value.IRecordValueDescriptorBuilder;
import de.xima.fc.interfaces.workflow.value.IStringValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptorFactory;
import lombok.SneakyThrows;

class OzgCloudPluginTest {

	@Spy
	@InjectMocks
	private OzgCloudPlugin plugin;

	@Mock
	private OzgPluginPropertiesSupplier propertiesMapper;
	@Mock
	private INodeExecutionParams<OzgPluginWorkflowNodeProperties> invocationParameters;

	@Nested
	@DisplayName("Success value descriptor")
	class TestSuccessValueDescriptor {

		@Mock
		private IValueDescriptorFactory descriptorFactory;
		@Mock
		private IRecordValueDescriptorBuilder descriptorBuilder;
		@Mock
		private IStringValueDescriptor stringValueDescriptor;

		@BeforeEach
		void setup() {
			when(descriptorFactory.recordBuilder()).thenReturn(descriptorBuilder);
			when(descriptorBuilder.requiredProperty(any(), any(IValueDescriptor.class))).thenReturn(descriptorBuilder);
		}

		@Test
		@DisplayName("should has vorgangnummer property")
		void shouldAddVorgangnummerProperty() {
			when(descriptorFactory.string()).thenReturn(stringValueDescriptor);

			plugin.getSuccessValueDescriptor(descriptorFactory);

			verify(descriptorBuilder).requiredProperty(VORGANGNUMMER_PROPERTY_KEY, stringValueDescriptor);
		}
	}

	@Nested
	class TestCreateOzgClient {

		@Mock
		private OzgPluginWorkflowNodeProperties formProperties;
		@Mock
		private NodeThrewExceptionFactory exceptionFactory;

		@Test
		@SneakyThrows
		void shouldCallBuildOzgCloudConfig() {
			when(invocationParameters.getData()).thenReturn(formProperties);
			doReturn("").when(plugin).getEingangsAdapterUrl(any());

			plugin.createOzgCloudClient(invocationParameters);

			verify(plugin).buildOzgCloudConfig(formProperties);
		}

		@Nested
		class TestEingangsAdapterUrl {

			private static final String CUSTOM_URL = "http://custom.url";

			@Test
			void shouldReturnCustomUrl() {
				when(formProperties.getEingangsAdapterUrl()).thenReturn(CUSTOM_URL);
				when(formProperties.isUseCustomEingangsAdapterUrl()).thenReturn(true);

				var eingangsAdapterUrl = plugin.getEingangsAdapterUrl(formProperties);

				assertThat(eingangsAdapterUrl).isEqualTo(CUSTOM_URL);
			}

			@Test
			void shouldReturnDefaultUrl() {
				when(formProperties.isUseCustomEingangsAdapterUrl()).thenReturn(false);
				when(propertiesMapper.getEingangsAdapterUrl()).thenReturn(PropertiesTestFactory.EINGANGSADAPTER_URL);

				var eingangsAdapterUrl = plugin.getEingangsAdapterUrl(formProperties);

				assertThat(eingangsAdapterUrl).isEqualTo(PropertiesTestFactory.EINGANGSADAPTER_URL);
			}

			@Test
			void shouldThrowExceptionIfNoUrl() {
				when(formProperties.isUseCustomEingangsAdapterUrl()).thenReturn(false);

				assertThrows(TechnicalException.class, () -> plugin.getEingangsAdapterUrl(formProperties));
			}
		}

		@Nested
		class TestBuildOzgClientConfig {

			@Mock
			private Properties properties;

			@Test
			void shouldSetEingangsAdapterUrl() {
				doReturn(PropertiesTestFactory.EINGANGSADAPTER_URL).when(plugin).getEingangsAdapterUrl(any());

				var ozgClientConfig = buildOzgCloudConfig();

				assertThat(ozgClientConfig.getEingangsAdapterUrl()).isEqualTo(PropertiesTestFactory.EINGANGSADAPTER_URL);
			}

			@SneakyThrows
			@Test
			void shouldSetProxyConfig() {
				var expectedProxyConfig = ProxyConfigTestFactory.create();
				when(propertiesMapper.getEingangsAdapterUrl()).thenReturn(PropertiesTestFactory.EINGANGSADAPTER_URL);
				when(propertiesMapper.getProxyProperties()).thenReturn(Optional.of(expectedProxyConfig));

				var ozgClientConfig = buildOzgCloudConfig();

				assertThat(ozgClientConfig.getProxyConfig()).isEqualTo(expectedProxyConfig);
			}

			@SneakyThrows
			private OzgCloudConfig buildOzgCloudConfig() {
				return plugin.buildOzgCloudConfig(formProperties);
			}
		}
	}

	@Nested
	class TestBuildSuccessResult {

		@Mock
		private INormalCompletionResultBuilder resultBuilder;
		@Captor
		private ArgumentCaptor<Map<String, String>> resultCaptor;

		@BeforeEach
		public void setup() {
			when(invocationParameters.normalResult()).thenReturn(resultBuilder);
			when(resultBuilder.success(resultCaptor.capture())).thenReturn(resultBuilder);
			when(resultBuilder.softError(any(), any(), any(), any())).thenReturn(resultBuilder);
		}

		@Test
		void shouldSetVorgangnummer() {
			plugin.buildSuccessResult(create(), invocationParameters);

			assertThat(resultCaptor.getValue()).containsExactly(Map.entry(VORGANGNUMMER_PROPERTY_KEY, VORGANGNUMMER));
		}

		@Test
		void shouldSetWarnings() {
			plugin.buildSuccessResult(create(), invocationParameters);

			verify(resultBuilder).softError(WARNING.getErrorCode(), WARNING.getMessage(), WARNING.getValue(), WARNING.getCause());
		}
	}

	@Nested
	class TestExecute {

		@Mock
		private OzgPluginExecutor pluginExecutor;
		@Mock
		private ExecutionResult executionResult;
		@Mock
		private INormalCompletionResult pluginCompletionResult;

		@BeforeEach
		void setup() {
			doReturn(pluginExecutor).when(plugin).createPluginExecutor(any());
		}

		@Test
		void shouldCallBuildSuccessResult() {
			when(pluginExecutor.execute()).thenReturn(executionResult);
			doReturn(pluginCompletionResult).when(plugin).buildSuccessResult(any(), any());

			executePlugin();

			verify(plugin).buildSuccessResult(executionResult, invocationParameters);
		}

		@Nested
		class TestCreateNodeThrewException {

			@Mock
			private NodeThrewExceptionFactory exceptionFactory;
			@Mock
			private OzgCloudRequestException ozgCloudRequestException;
			@Mock
			private TechnicalException technicalException;
			@Mock
			private RuntimeException unexpectedException;
			@Mock
			private NodeThrewException nodeThrewException;

			@BeforeEach
			void setup() {
				doReturn(exceptionFactory).when(plugin).createExceptionFactory(any());
			}

			@Test
			void shouldCallCreateOzgCloudRequestException() {
				when(pluginExecutor.execute()).thenThrow(ozgCloudRequestException);
				when(exceptionFactory.createOzgCloudRequestException(any())).thenReturn(nodeThrewException);

				executePlugin();

				verify(exceptionFactory).createOzgCloudRequestException(ozgCloudRequestException);
			}

			@Test
			void shouldCallCreateInternalException() {
				when(pluginExecutor.execute()).thenThrow(technicalException);
				when(exceptionFactory.createInternalException(any())).thenReturn(nodeThrewException);

				executePlugin();

				verify(exceptionFactory).createInternalException(technicalException);
			}

			@Test
			void shouldCallCreateInternalExceptionWhenUnexpectedThrown() {
				when(pluginExecutor.execute()).thenThrow(unexpectedException);
				when(exceptionFactory.createInternalException(any(), any())).thenReturn(nodeThrewException);

				executePlugin();

				verify(exceptionFactory).createInternalException(any(), eq(unexpectedException));
			}
		}

		void executePlugin() {
			try {
				plugin.execute(invocationParameters);
			} catch (Exception e) {
			}
		}

	}
}