package de.ozgcloud.formcycle.errorhandling;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.xima.fc.interfaces.workflow.params.INodeThrewExceptionBuilder;

class NodeThrewExceptionFactoryTest {

	private static final String EXCEPTION_ID = "exception-id";

	@Spy
	@InjectMocks
	private NodeThrewExceptionFactory exceptionFactory;

	@Mock
	private INodeThrewExceptionBuilder exceptionBuilder;

	@Captor
	private ArgumentCaptor<String> stringCaptor;
	@Captor
	private ArgumentCaptor<Map<String, String>> errorDataCaptor;
	@Captor
	private ArgumentCaptor<Exception> errorCauseCaptor;
	@Mock
	private FormcycleAttachment expectedAttachments;

	@BeforeEach
	void setup() {
		when(exceptionBuilder.error(any(), any())).thenReturn(exceptionBuilder);
		when(exceptionBuilder.cause(any())).thenReturn(exceptionBuilder);
	}

	@DisplayName("Create internal exception")
	@Nested
	class TestCreateInternalException {

		@DisplayName("form TechnicalException")
		@Nested
		class TestFromTechnicalException {
			@Mock
			private TechnicalException exception;

			@BeforeEach
			void setup() {
				when(exception.getExceptionId()).thenReturn(EXCEPTION_ID);
			}

			@Test
			void shouldAddErrorCode() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.error(stringCaptor.capture(), any())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.INTERNAL_ERROR.name());
			}

			@Test
			void shouldAddExceptionId() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.error(any(), errorDataCaptor.capture())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				var errorData = errorDataCaptor.getValue();
				assertThat(errorData).containsEntry(NodeThrewExceptionFactory.EXCEPTION_ID_KEY, EXCEPTION_ID);
			}

			@Test
			void shouldAddExceptionMessage() {
				var errorMessage = "error message";
				when(exception.getMessage()).thenReturn(errorMessage);
				when(exceptionBuilder.message(stringCaptor.capture())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				assertThat(stringCaptor.getValue()).isEqualTo(errorMessage);
			}

			@Test
			void shouldAddCause() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.cause(errorCauseCaptor.capture())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
			}
		}

		@DisplayName("from message and exception")
		@Nested
		class TestFromException {

			private static final String EXCEPTION_MESSAGE = "exception message";
			@Mock
			private Exception exception;

			@Test
			void shouldAddErrorCode() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.error(stringCaptor.capture(), any())).thenReturn(exceptionBuilder);

				createException();

				assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.INTERNAL_ERROR.name());
			}

			@Test
			void shouldAddExceptionId() {
				doReturn(Map.of(NodeThrewExceptionFactory.EXCEPTION_ID_KEY, EXCEPTION_ID)).when(exceptionFactory).createDataMap(any());
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.error(any(), errorDataCaptor.capture())).thenReturn(exceptionBuilder);

				createException();

				var errorData = errorDataCaptor.getValue();
				assertThat(errorData).containsEntry(NodeThrewExceptionFactory.EXCEPTION_ID_KEY, EXCEPTION_ID);
			}

			@Test
			void shouldAddExceptionMessage() {
				when(exceptionBuilder.message(stringCaptor.capture())).thenReturn(exceptionBuilder);

				createException();

				assertThat(stringCaptor.getValue()).contains(EXCEPTION_MESSAGE, "(ExceptionId:");
			}

			@Test
			void shouldAddCause() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exceptionBuilder.cause(errorCauseCaptor.capture())).thenReturn(exceptionBuilder);

				createException();

				assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
			}

			private void createException() {
				exceptionFactory.createInternalException(EXCEPTION_MESSAGE, exception);
			}
		}
	}

	@Nested
	class TestFromOzgCloudRequestException {
		@Mock
		private OzgCloudRequestException exception;

		@BeforeEach
		void setup() {
			when(exception.getExceptionId()).thenReturn(EXCEPTION_ID);
		}

		@Test
		void shouldAddErrorCode() {
			when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.error(stringCaptor.capture(), any())).thenReturn(exceptionBuilder);

			createException();

			assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.OZGCLOUD_REQUEST_FAILED.name());
		}

		@Test
		void shouldAddExceptionId() {
			when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.error(any(), errorDataCaptor.capture())).thenReturn(exceptionBuilder);

			createException();

			var errorData = errorDataCaptor.getValue();
			assertThat(errorData).containsEntry(NodeThrewExceptionFactory.EXCEPTION_ID_KEY, EXCEPTION_ID);
		}

		@Test
		void shouldAddExceptionMessage() {
			var errorMessage = "error message";
			when(exception.getMessageWithConfig()).thenReturn(errorMessage);
			when(exceptionBuilder.message(stringCaptor.capture())).thenReturn(exceptionBuilder);

			createException();

			assertThat(stringCaptor.getValue()).isEqualTo(errorMessage);
		}

		@Test
		void shouldAddCause() {
			when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.cause(errorCauseCaptor.capture())).thenReturn(exceptionBuilder);

			createException();

			assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
		}

		private void createException() {
			exceptionFactory.createOzgCloudRequestException(exception);
		}
	}
}