/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.formdata;

import static de.ozgcloud.formcycle.formdata.FormDataTestFactory.*;
import static de.ozgcloud.formcycle.formdata.PluginDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.xima.fc.entities.Projekt;
import de.xima.fc.form.common.models.XForm;
import de.xima.fc.interfaces.form.IFormDataAdapter;
import de.xima.fc.interfaces.workflow.execution.IWorkflowExecutionEnvironmentData;
import de.xima.fc.mdl.enums.EFormFieldTyp;
import de.xima.fc.mdl.fdv.form.FormFieldMetaData;

class PluginFormDataAdapterTest {

	@Spy
	@InjectMocks
	private PluginFormDataAdapter adapter;

	@Mock
	private IWorkflowExecutionEnvironmentData environmentData;
	@Mock
	private IFormDataAdapter formDataAdapter;
	@Mock
	private FormFieldMetaData formFieldMetaData;
	@Mock
	private Projekt projekt;
	@Mock
	private XForm xForm;


	@Nested
	@DisplayName("FormDataAdapter")
	class TestFormDataAdapter {

		@BeforeEach
		void setup() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Nested
		@DisplayName("Read Form Data")
		class TestReadData {

			@BeforeEach
			void setup() {
				when(projekt.getTitel()).thenReturn(FORM_NAME);
				when(environmentData.getProject()).thenReturn(projekt);
				when(environmentData.getXForm()).thenReturn(xForm);
				var xItems = StructureMockFactory.create();
				when(xForm.getXItems()).thenReturn(xItems);
				adapter = spy(new PluginFormDataAdapter(environmentData, ORGANISATIONS_EINHEIT_ID));
			}

			@Test
			void shouldSetOrganisationseinheitenId() {
				var formData = adapter.readFormData();

				assertThat(formData.getOrganisationsEinheitId()).isEqualTo(ORGANISATIONS_EINHEIT_ID);
			}

			@Test
			void shouldSetFormName() {
				var formData = adapter.readFormData();

				assertThat(formData.getFormName()).isEqualTo(FORM_NAME);
			}

			@Test
			void shouldCallAddFormData() {
				adapter.readFormData();

				verify(adapter).buildFormData();
			}

			@Test
			void shouldCallAddAttachments() {
				adapter.readFormData();

				verify(adapter).buildFormData();
			}
		}

		@Nested
		class TestGetValueAsText {

			@Nested
			class TestSelectFormFields {

				@BeforeEach
				void setup() {
					when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.SELECT);
					when(formDataAdapter.getFieldMetaData(anyString(), anyBoolean())).thenReturn(formFieldMetaData);
				}

				@Test
				void shouldReturnExpectedValue() {
					when(formFieldMetaData.getOptionsMap()).thenReturn(createOptionsMap());

					var valueAsText = getValueAsText(OPTIONS_KEY);

					assertThat(valueAsText).isEqualTo(TEXT_MAP_VALUE);
				}

				@Test
				void shouldReturnInputWhenNoOption() {
					when(formFieldMetaData.getOptionsMap()).thenReturn(new HashMap<>());

					var result = getValueAsText(FIELD_VALUE);

					assertThat(result).isEqualTo(FIELD_VALUE);

				}

				@Test
				void shouldReturnWhenNoTextValue() {
					when(formFieldMetaData.getOptionsMap()).thenReturn(createOptionsMapWithTextMapKey(TITLE_MAP_KEY));

					var result = getValueAsText(FIELD_VALUE);

					assertThat(result).isEqualTo(FIELD_VALUE);
				}

			}
		}

		@Nested
		class TestGetValues {

			@Test
			void shouldCallConvertMethod() {
				var values1 = new String[] { "0", "1" };
				var values2 = new String[] { "2" };
				when(formDataAdapter.getFieldValues(anyString())).thenReturn(List.of(values1, values2));

				adapter.getValues(FIELD_KEY);

				verify(adapter).getValueAsText(FIELD_KEY, "0");
				verify(adapter).getValueAsText(FIELD_KEY, "1");
				verify(adapter).getValueAsText(FIELD_KEY, "2");
			}
		}

		@Nested
		class TestAddFormData {

			@BeforeEach
			void setup() {
				when(environmentData.getXForm()).thenReturn(xForm);
				var xItems = StructureMockFactory.create();
				when(xForm.getXItems()).thenReturn(xItems);
			}

			@Test
			@DisplayName("should call getValues when converting form data")
			void shouldCallGetValues() {
				var xItem = StructureMockFactory.createItem(ItemClass.TEXT_FIELD);
				when(xForm.getXItems()).thenReturn(xItem);

				adapter.buildFormData();

				verify(adapter).getValues(StructureMockFactory.ITEM_NAME);
			}

			@Test
			@DisplayName("should call setFieldData when converting form data")
			void shouldCallSetFieldData() {
				var formData = adapter.buildFormData();

				assertThat(formData).isNotEmpty();
			}

		}

		@Nested
		class TestGetUploadFields {

			@Test
			@DisplayName("should skip not upload fields")
			void shouldSkipFields() {
				when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.TEXTFIELD);
				when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Map.of(FIELD_KEY, formFieldMetaData));

				var uploadFieldNames = adapter.getUploadFieldNames();

				assertThat(uploadFieldNames).isEmpty();
			}

			@Test
			void shouldReturnWhenNoData() {
				when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Collections.emptyMap());

				var uploadFieldNames = adapter.getUploadFieldNames();

				assertThat(uploadFieldNames).isEmpty();
			}

			@Test
			@DisplayName("should return names of upload fields")
			void shouldReturnFieldNames() {
				when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.UPLOAD);
				when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Map.of(FIELD_KEY, formFieldMetaData));

				var uploadFieldNames = adapter.getUploadFieldNames();

				assertThat(uploadFieldNames).containsExactlyInAnyOrder(FIELD_KEY);
			}
		}

		@Nested
		@DisplayName("Get Attachment Names")
		class TestAddAttachmentNames {

			private static final String FIELD_VALUE_1 = "value1";
			private static final String FIELD_VALUE_2 = "value2";

			@Test
			@DisplayName("should skip upload field when no content was uploaded")
			void shouldSkipWhenNoData() {
				when(formDataAdapter.getFieldValues(anyString())).thenReturn(null);

				var attachmentNames = adapter.getAttachmentNames(Set.of(FIELD_KEY));

				assertThat(attachmentNames).isEmpty();
			}

			@Test
			@DisplayName("should add all attachments when multiple elements have the same name")
			void shouldGetAllAttachments1() {
				when(formDataAdapter.getFieldValues(anyString())).thenReturn(createFieldValuesWhenEqualUploadName(FIELD_VALUE_1, FIELD_VALUE_2));

				var attachmentNames = adapter.getAttachmentNames(Set.of(FIELD_KEY));

				assertThat(attachmentNames).containsExactlyInAnyOrder(FIELD_VALUE_1, FIELD_VALUE_2);
			}

			@Test
			@DisplayName("should add all attachments")
			void shouldGetAllAttachments2() {
				when(formDataAdapter.getFieldValues(anyString())).thenReturn(createFieldValues(FIELD_VALUE_1, FIELD_VALUE_2));

				var attachmentNames = adapter.getAttachmentNames(Set.of(FIELD_KEY));

				assertThat(attachmentNames).containsExactlyInAnyOrder(FIELD_VALUE_1, FIELD_VALUE_2);
			}
		}

		@Nested
		@DisplayName("Get Attachment UUIDs")
		class TestAttachmentUuids {

			private static final String FIELD_VALUE_1 = "value1";

			private Map<String, List<String[]>> fieldValuesMap = new HashMap<>();

			@BeforeEach
			void setup() {
				fieldValuesMap.put(FIELD_KEY, createFieldValues(FIELD_VALUE_1));
				fieldValuesMap.put(UUID_FIELD_KEY, createFieldValues(UUID_VALUE));
				when(formDataAdapter.getFieldValuesMap()).thenReturn(fieldValuesMap);

			}

			@Test
			void shouldGetAttachmentUuids() {
				var attachmentUuids = adapter.getAttachmentUuids(Set.of(FIELD_KEY));

				assertThat(attachmentUuids).hasSize(1).first().isEqualTo(UUID_VALUE);
			}

			@Test
			@DisplayName("should add all attachments when upload elements have different names")
			void shouldGetAllAttachmentUuidsFromList() {
				var uuid2 = UUID.randomUUID().toString();
				fieldValuesMap.put("field2.uuid", createFieldValues(UUID_VALUE, uuid2));

				var attachmentUuids = adapter.getAttachmentUuids(Set.of(FIELD_KEY, "field2"));

				assertThat(attachmentUuids).hasSize(2).containsExactlyInAnyOrder(UUID_VALUE, uuid2);
			}

			@Test
			@DisplayName("should add all attachments when upload elements have same name")
			void shouldGetAllAttachmentUuidsFromArray() {
				var uuid2 = UUID.randomUUID().toString();
				fieldValuesMap.put("field2.uuid", createFieldValuesWhenEqualUploadName(UUID_VALUE, uuid2));

				var attachmentUuids = adapter.getAttachmentUuids(Set.of(FIELD_KEY, "field2"));

				assertThat(attachmentUuids).hasSize(2).containsExactlyInAnyOrder(UUID_VALUE, uuid2);
			}
		}
	}

	@Nested
	@DisplayName("Get value as text")
	class TestEdgeCases {

		@Test
		@DisplayName("should return same value when field name is null")
		void shouldReturnValueWhenNullFieldName() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);

			var result = adapter.getValueAsText(null, FIELD_VALUE);

			assertThat(result).isEqualTo(FIELD_VALUE);
		}

		@Test
		@DisplayName("should return null when field value is null")
		void shouldReturnWhenValueNull() {
			var result = getValueAsText(null);

			assertThat(result).isNull();
		}

		@Test
		@DisplayName("should return same value when field value is blank")
		void shouldReturnWhenValueBlank() {
			var result = getValueAsText(StringUtils.SPACE);

			assertThat(result).isEqualTo(StringUtils.SPACE);
		}

		@Test
		@DisplayName("should return same value when field value is empty")
		void shouldReturnWhenValueEmpty() {
			var result = getValueAsText(StringUtils.EMPTY);

			assertThat(result).isEmpty();
		}

		@Test
		@DisplayName("should return same value when there's no metadata")
		void shouldReturnWhenNoMetadata() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);

			var result = getValueAsText(FIELD_VALUE);

			assertThat(result).isEqualTo(FIELD_VALUE);

		}

		@Test
		@DisplayName("should return same value when element type isn't SELECT")
		void shouldReturnWhenNotSelect() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.TEXTFIELD);
			when(formDataAdapter.getFieldMetaData(anyString(), anyBoolean())).thenReturn(formFieldMetaData);

			var result = getValueAsText(FIELD_VALUE);

			assertThat(result).isEqualTo(FIELD_VALUE);
		}
	}

	private String getValueAsText(String value) {
		return adapter.getValueAsText(FIELD_KEY, value);
	}

	@Nested
	@DisplayName("Set PostkorbId")
	class TestPostkorbId {

		@BeforeEach
		void setup() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		void shouldCallGetPostkorbId() {
			when(environmentData.getProject()).thenReturn(projekt);
			when(environmentData.getXForm()).thenReturn(xForm);
			var xItems = StructureMockFactory.create();
			when(xForm.getXItems()).thenReturn(xItems);

			adapter.readFormData();

			verify(adapter).getPostkorbId();
		}

		@Test
		void shouldMapPostkorbId() {
			var values = new ArrayList<String[]>();
			values.add(new String[]{ POSTKORB_ID });
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(values);

			Optional<String> postkorbId = adapter.getPostkorbId();

			assertThat(postkorbId).contains(POSTKORB_ID);
		}

		@Test
		void shouldNotMapOnEmptyPostkorbId() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(Collections.emptyList());

			Optional<String> postkorbId = adapter.getPostkorbId();

			assertThat(postkorbId).isEmpty();
		}

	}

	@Nested
	class TestPostkorbIdValidation {

		@DisplayName("should return FALSE when")
		@ParameterizedTest(name = "postkorbId is \"{0}\"")
		@NullSource
		@ValueSource(strings = {"", "[%$id%]", "[% $id%]", "[%$id %]", "[% $id %]"})
		void shouldFilterOutEmptyValues(String postkorbId) {
			var isValid = adapter.isValidPostkorbId(postkorbId);

			assertThat(isValid).isFalse();
		}
	}

	@Nested
	@DisplayName("Set PostkorbId")
	class TestCleanFormFields {

		@Test
		void shouldCallGetPostkorbId() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			when(environmentData.getProject()).thenReturn(projekt);
			when(environmentData.getXForm()).thenReturn(xForm);
			var xItems = StructureMockFactory.create();
			when(xForm.getXItems()).thenReturn(xItems);

			adapter.readFormData();

			verify(adapter).getPostkorbId();
		}

		@Test
		void shouldRemovePostkorbIdNode() {
			var formNode = FormNodeTestFactory.createBuilder().name(PluginFormDataAdapter.POSTKORB_FORM_KEY).build();

			var formFieldsOut = adapter.setSubNodeValues(List.of(formNode));

			assertThat(formFieldsOut).isEmpty();
		}

		@Test
		void shouldNotRemoveOtherNodes() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			FormNode formNode = FormNodeTestFactory.create();

			var formNodes = adapter.setSubNodeValues(List.of(formNode));

			assertThat(formNodes).hasSize(1);
			verify(adapter).isPostkorbIdNode(any());
		}

	}
}